"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Via = void 0;
const rowan_1 = require("rowan");
const events_1 = require("events");
const core_1 = require("@viae/core");
const context_1 = require("./context");
const data_decoder_1 = require("./middleware/data-decoder");
const data_encoder_1 = require("./middleware/data-encoder");
const interceptor_1 = require("./middleware/interceptor");
const send_1 = require("./middleware/send");
const pb_1 = require("@viae/pb");
const pino_1 = require("pino");
const util_1 = require("./util");
const readable_stream_1 = require("./middleware/readable-stream");
const normalise_1 = require("./util/normalise");
/**
 * Via
 * Wraps a wire connection and processes inbound and outbound messages
 */
class Via extends rowan_1.Rowan {
    _ev = new events_1.EventEmitter();
    _active = [];
    _wire;
    _uuid;
    _log;
    _timeout;
    _interceptor;
    _encoder = new pb_1.FrameEncoder();
    _before = new rowan_1.Rowan();
    CtxCtor;
    /* outgoing message pipeline */
    out = new rowan_1.Rowan();
    get wire() {
        return this._wire;
    }
    get active() {
        return this._active;
    }
    static Log = (0, pino_1.pino)();
    constructor(opts) {
        super();
        const wire = this._wire = opts.wire;
        const codex = (opts ? opts.codex : undefined) || util_1.Codex.createDefault();
        this.CtxCtor = (opts ? opts.Ctx : undefined) || context_1.DefaultContext;
        this._log = (opts ? opts.log : undefined) || Via.Log;
        this._interceptor = new interceptor_1.default(this._log);
        this._uuid = (opts ? opts.uuid : undefined) || util_1.shortId;
        this._timeout = (opts ? opts.timeout : undefined) || 60000;
        this.meta = {
            type: "Via"
        };
        const self = this;
        this
            /* execute the 'before' pipeline */
            .use(this._before)
            /* execute the 'outbound' pipeline after the main pipeline */
            .use(new rowan_1.After([
            this.out
                .use(new rowan_1.AfterIf(function (ctx) { return Promise.resolve(!!ctx.out); }, [
                new readable_stream_1.UpgradeOutgoingReadableStream(),
                new data_encoder_1.default(codex),
                new send_1.default(this._encoder)
            ]))
        ]))
            .use(new rowan_1.Catch(function (err, ctx) {
            ctx.err = err;
            self._log.error({ err }, "error during processing");
            if (ctx.out) {
                ctx.out.head.status = core_1.Status.Error;
                ctx.out.data = err.message;
            }
            return Promise.resolve();
        }))
            /* add the lazy data decoder */
            .use(new data_decoder_1.default(codex))
            /* convert data into readable-stream  */
            .use(new readable_stream_1.UpgradeIncomingReadableStream())
            /* intercept id-matching messages */
            .use(this._interceptor);
        /* ... then any more .use() middleware
        
        /** Hook onto the wire events*/
        wire.on("message", (data) => {
            this._onMessage(data);
        });
        wire.on("open", () => {
            this._ev.emit("open");
        });
        wire.on("close", () => {
            this._interceptor.dispose();
            this._ev.emit("close");
        });
        wire.on("error", (err) => {
            this._ev.emit("error", err);
        });
    }
    get log() {
        return this._log;
    }
    get ready() {
        return new Promise((r, x) => {
            if (this._wire.readyState == core_1.WireState.OPEN) {
                return r();
            }
            const handlerOpen = () => {
                this._wire.off("open", handlerOpen);
                this._wire.off("error", handlerError);
                r();
            };
            const handlerError = (err) => {
                this._wire.off("open", handlerOpen);
                this._wire.off("error", handlerError);
                x(err);
            };
            this._wire.on("open", handlerOpen);
            this._wire.on("error", handlerError);
        });
    }
    _onMessage(data) {
        const msg = this._encoder.decode((0, util_1.toUint8Array)(data));
        const ctx = new this.CtxCtor({ connection: this, in: msg, log: this._log });
        this._active.push(ctx);
        this.process(ctx)
            .then(() => {
            return ctx.complete;
        })
            .catch(err => {
            this._log.error({ err, ctx }, "unhandled error");
            this._ev.emit("error", err);
        }).finally(() => {
            let index = this._active.indexOf(ctx);
            this._active.splice(index, 1);
            return ctx[Symbol.asyncDispose]();
        });
    }
    /**
     * Fire and Forget - Add a message to the outgoing pipeline
     **/
    send(msg, opts) {
        if (!msg.id)
            msg.id = (0, util_1.shortId)();
        if (opts && opts.encoding) {
            msg.head = msg.head || {};
            msg.head.encoding = opts.encoding;
        }
        if (opts && typeof (opts.head) == "object") {
            Object.assign(msg.head, opts.head);
        }
        if (this._wire == undefined || this._wire.readyState !== core_1.WireState.OPEN) {
            throw Error("wire is not open");
        }
        let ctx = new this.CtxCtor({ connection: this, out: msg, log: this._log });
        return this.out.process(ctx)
            .then(() => ctx.complete)
            .catch((err) => {
            this._ev.emit("error", err);
            throw err;
        })
            .finally(() => {
            return ctx[Symbol.asyncDispose]();
        });
    }
    /**
     * Send a request message, add an one-off interceptor for the response. Returns promise to first response.
     * @param msg
     * @param opts
     */
    async request(method, path, data, opts) {
        path = (0, normalise_1.normalisePath)(path);
        let msg = {
            id: (opts && opts.id) ? opts.id : (0, util_1.shortId)(),
            head: { method, path }
        };
        if (data) {
            msg.data = data;
        }
        let reject, resolve, promise = new Promise((r, x) => { resolve = r; reject = x; });
        let clock;
        let dispose = this._interceptor.intercept({
            id: msg.id,
            handlers: [function (ctx, _) {
                    const status = ctx.in.head.status;
                    if (status == core_1.Status.OkayPartial) {
                        clearTimeout(clock);
                        clock = setTimeout(() => { reject("Request Timeout"); }, 2 * 60 * 1000);
                        return;
                    }
                    resolve({
                        ok: status >= 200 && status < 300,
                        get data() { return ctx.in.data; },
                        [Symbol.asyncDispose]() {
                            return ctx.complete.then(_ => ctx[Symbol.asyncDispose]());
                        },
                        res: ctx.in
                    });
                    dispose();
                    return Promise.resolve();
                }]
        });
        clock = setTimeout(() => { reject("Request Timeout"); }, (opts ? opts.timeout : undefined) || this._timeout);
        try {
            await this.send(msg, opts);
            return await promise;
        }
        catch (err) {
            throw err;
        }
        finally {
            clearTimeout(clock);
            dispose();
        }
    }
    /**
     * simplified request/response  - this will deserialize the response data and return it if successful.
     * throws an error if the result status is not OK.
     **/
    /*async call<E = any>(opts: CallOptions<E>): Promise<E> {
      const { method, path, data } = opts;
      let result = await this.request(method, path, data, opts);
      switch (Number(result.head.status)) {
        case Status.NotFound:
          throw Error(`${method} "${path}" not found`);
        case Status.Unauthorized:
          throw Error(`${method} "${path}" not authorized`);
        case Status.Forbidden:
          throw Error(`${method} "${path}" forbidden`);
        case Status.BadRequest:
          throw Error(`${method} "${path}" bad request`);
        case Status.Error:
          let msg = `${method} "${path}" error`;
          let resultData = result.data;
          if (resultData != undefined && typeof resultData === "string") {
            msg += ": " + resultData;
          }
          throw Error(msg);
        case Status.OK:
        case Status.OkayPartial:
          let validate = (opts && typeof opts.validate == "function") ? opts.validate : () => true;
          validate(result.data);
          return result.data;
        default:
          throw Error(`${method} "${path}" unknown status code: ${result.head.status}`)
      }
    }*/
    /**
     * intercept an id-specific message
     * @param id message id to intercept
     * @param handlers processors to call on interception
     */
    intercept(id, handlers) {
        return this._interceptor.intercept({ id: id, handlers: handlers });
    }
    /**
     * processors to use before any  in-built (inbound)processing
     * @param processor
     */
    before(processor) {
        this._before.use(processor);
        return this;
    }
    createId() {
        return this._uuid();
    }
    on(event, cb) {
        this._ev.on(event, cb);
    }
    off(event, cb) {
        this._ev.off(event, cb);
    }
}
exports.Via = Via;
